/*
 * Copyright 2020 Google LLC
 * Copyright 2020 EPAM Systems, Inc
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     https://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.opengroup.osdu.indexer.service;

import com.google.gson.Gson;
import org.apache.http.StatusLine;
import org.elasticsearch.ElasticsearchException;
import org.elasticsearch.action.admin.indices.mapping.get.GetFieldMappingsResponse;
import org.elasticsearch.action.admin.indices.mapping.get.GetFieldMappingsResponse.FieldMappingMetaData;
import org.elasticsearch.action.bulk.BulkItemResponse.Failure;
import org.elasticsearch.action.support.master.AcknowledgedResponse;
import org.elasticsearch.client.*;
import org.elasticsearch.common.bytes.BytesReference;
import org.elasticsearch.common.xcontent.XContentBuilder;
import org.elasticsearch.common.xcontent.XContentFactory;
import org.elasticsearch.index.reindex.BulkByScrollResponse;
import org.junit.Before;
import org.junit.Ignore;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.common.model.indexer.IndexSchema;
import org.opengroup.osdu.core.common.model.search.RecordMetaAttribute;
import org.opengroup.osdu.indexer.config.IndexerConfigurationProperties;
import org.opengroup.osdu.indexer.util.ElasticClientHandler;
import org.opengroup.osdu.indexer.util.TypeMapper;
import org.powermock.api.mockito.PowerMockito;
import org.powermock.core.classloader.annotations.PrepareForTest;
import org.springframework.test.context.junit4.SpringRunner;

import java.io.IOException;
import java.util.*;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.fail;
import static org.mockito.Matchers.any;
import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.mock;
import static org.mockito.MockitoAnnotations.initMocks;
import static org.powermock.api.mockito.PowerMockito.when;

@Ignore
@RunWith(SpringRunner.class)
@PrepareForTest({ RestHighLevelClient.class, IndicesClient.class})
public class IndexerMappingServiceTest {

	private final String kind = "tenant:test:test:1.0.0";
	private final String index = "tenant-test-test-1.0.0";
	private final String type = "test";
	private final String mappingValid = "{\"dynamic\":false,\"properties\":{\"data\":{\"properties\":{\"Msg\":{\"type\":\"text\",\"analyzer\":\"de_indexer_analyzer\",\"search_analyzer\":\"de_search_analyzer\"},\"Location\":{\"type\":\"geo_point\"}}},\"id\":{\"type\":\"keyword\"},\"acl\":{\"properties\":{\"viewers\":{\"type\":\"keyword\"},\"owners\":{\"type\":\"keyword\"}}}}}";

	@Mock
	private IndexerConfigurationProperties configurationProperties;
	@Mock
	private RestClient restClient;
	@Mock
	private Response response;
	@Mock
	private StatusLine statusLine;

	@InjectMocks
	private IndexerMappingServiceImpl sut;

	@Mock
    private ElasticClientHandler elasticClientHandler;

	@InjectMocks
	private RestHighLevelClient restHighLevelClient;

	@InjectMocks
	private IndexSchema indexSchema;
	@InjectMocks
	private IndicesClient indicesClient;

	@InjectMocks
	private AcknowledgedResponse mappingResponse;

	@Before
	public void setup() throws IOException {
		initMocks(this);
		when(configurationProperties.isPreDemo()).thenReturn(true);
		Map<String, String> dataMapping = new HashMap<>();
		dataMapping.put("Location", "geo_point");
		dataMapping.put("Msg", "text");
		Map<String, Object> metaMapping = new HashMap<>();
		metaMapping.put(RecordMetaAttribute.ID.getValue(), "keyword");
		metaMapping.put(RecordMetaAttribute.ACL.getValue(), TypeMapper.getIndexerType(RecordMetaAttribute.ACL));
		this.indexSchema = IndexSchema.builder().kind(kind).type(type).dataSchema(dataMapping).metaSchema(metaMapping)
				.build();

		this.indicesClient = PowerMockito.mock(IndicesClient.class);
		this.restHighLevelClient = PowerMockito.mock(RestHighLevelClient.class);

		when(this.restHighLevelClient.getLowLevelClient()).thenReturn(restClient);
		when(this.restClient.performRequest(any())).thenReturn(response);
		when(this.response.getStatusLine()).thenReturn(statusLine);
		when(this.statusLine.getStatusCode()).thenReturn(200);
	}

	@Test
	public void should_returnValidMapping_givenFalseMerge_createMappingTest() {
		try {
			String mapping = this.sut.createMapping(restHighLevelClient, indexSchema, index, false);
			assertEquals(mappingValid, mapping);
		} catch (Exception e) {
			fail("Should not throw this exception" + e.getMessage());
		}
	}

	@Test
	public void should_returnValidMapping_givenTrueMerge_createMappingTest() {
		try {
			doReturn(this.indicesClient).when(this.restHighLevelClient).indices();
			doReturn(mappingResponse).when(this.indicesClient).putMapping(any(), any(RequestOptions.class));

			String mapping = this.sut.createMapping(this.restHighLevelClient, this.indexSchema, this.index, true);
			assertEquals(this.mappingValid, mapping);
		} catch (Exception e) {
			fail("Should not throw this exception" + e.getMessage());
		}
	}

	@Test
	public void should_returnValidMapping_givenExistType_createMappingTest() {
		try {
			doReturn(this.indicesClient).when(this.restHighLevelClient).indices();
			doReturn(mappingResponse).when(this.indicesClient).putMapping(any(), any(RequestOptions.class));

			IndexerMappingServiceImpl indexerMappingServiceLocal = PowerMockito.spy(new IndexerMappingServiceImpl());
			doReturn(false).when(indexerMappingServiceLocal).isTypeExist(any(), any(), any());
			String mapping = this.sut.createMapping(this.restHighLevelClient, this.indexSchema, this.index, true);
			assertEquals(this.mappingValid, mapping);
		} catch (Exception e) {
			fail("Should not throw this exception" + e.getMessage());
		}
	}

	@Test
	public void should_update_indices_field_with_keyword_when_valid_indices() throws Exception {
		try {
			Set<String> indices = new HashSet<String>();
			indices.add("indices 1");
			GetFieldMappingsResponse getFieldMappingsResponse = mock(GetFieldMappingsResponse.class);
			doReturn(this.indicesClient).when(this.restHighLevelClient).indices();
			when(this.indicesClient.getFieldMapping(any(), any())).thenReturn(getFieldMappingsResponse);
			XContentBuilder builder = XContentFactory.jsonBuilder();
			builder.startObject();
			builder.field("any field", new HashMap());
			builder.endObject();
			BytesReference bytesReference = BytesReference.bytes(builder);
			FieldMappingMetaData mappingMetaData = new FieldMappingMetaData(index, bytesReference);
			Map<String, FieldMappingMetaData> mapBuilder = new HashMap<>();
			mapBuilder.put("data.any field", mappingMetaData);
			Map<String, Map<String, FieldMappingMetaData>> mappingBuilder = new HashMap<>();
			mappingBuilder.put("any index 1", mapBuilder);
			mappingBuilder.put("any index 2", mapBuilder);
			Map<String, Map<String, Map<String, FieldMappingMetaData>>> mapping = new HashMap<>();
			mapping.put("indices 1", mappingBuilder);
			when(getFieldMappingsResponse.mappings()).thenReturn(mapping);
			doReturn(mappingResponse).when(this.indicesClient).putMapping(any(), any(RequestOptions.class));
			BulkByScrollResponse response = mock(BulkByScrollResponse.class);
			doReturn(response).when(this.restHighLevelClient).updateByQuery(any(), any(RequestOptions.class));
			when(response.getBulkFailures()).thenReturn(new ArrayList<Failure>());
			when(elasticClientHandler.createRestClient()).thenReturn(restHighLevelClient);

			this.sut.updateIndexMappingForIndicesOfSameType( indices,"any field");
		} catch (Exception e) {
			fail("Should not throw this exception" + e.getMessage());
		}
	}

	@Test(expected = AppException.class)
	public void should_throw_exception_if_someIndex_is_invalid_andWeIndexfield_with_keyword() throws Exception {
		try {
			Set<String> indices = new HashSet<String>();
			indices.add("invalid 1");
			GetFieldMappingsResponse getFieldMappingsResponse = mock(GetFieldMappingsResponse.class);
			doReturn(this.indicesClient).when(this.restHighLevelClient).indices();
			when(this.indicesClient.getFieldMapping(any(), any())).thenReturn(getFieldMappingsResponse);
			XContentBuilder builder = XContentFactory.jsonBuilder();
			builder.startObject();
			builder.field("any field", new HashMap());
			builder.endObject();
			BytesReference bytesReference = BytesReference.bytes(builder);
			FieldMappingMetaData mappingMetaData = new FieldMappingMetaData(index, bytesReference);
			Map<String, FieldMappingMetaData> mapBuilder = new HashMap<>();
			mapBuilder.put("data.any field", mappingMetaData);
			Map<String, Map<String, FieldMappingMetaData>> mappingBuilder = new HashMap<>();
			mappingBuilder.put("any index 1", mapBuilder);
			mappingBuilder.put("any index 2", mapBuilder);
			Map<String, Map<String, Map<String, FieldMappingMetaData>>> mapping = new HashMap<>();
			mapping.put("indices 1", mappingBuilder);
			when(getFieldMappingsResponse.mappings()).thenReturn(mapping);
			doReturn(mappingResponse).when(this.indicesClient).putMapping(any(), any(RequestOptions.class));
			BulkByScrollResponse response = mock(BulkByScrollResponse.class);
			doReturn(response).when(this.restHighLevelClient).updateByQuery(any(), any(RequestOptions.class));
			when(response.getBulkFailures()).thenReturn(new ArrayList<Failure>());
			when(elasticClientHandler.createRestClient()).thenReturn(restHighLevelClient);

			this.sut.updateIndexMappingForIndicesOfSameType(indices,"any field");
		} catch (Exception e) {
			throw e;
		}
	}

	@Test(expected = AppException.class)
	public void should_throw_exception_if_type_of_index_is_invalid_andWeIndexfield_with_keyword() throws Exception {
		try {
			Set<String> indices = new HashSet<String>();
			indices.add("indices 1");
			GetFieldMappingsResponse getFieldMappingsResponse = mock(GetFieldMappingsResponse.class);
			doReturn(this.indicesClient).when(this.restHighLevelClient).indices();
			when(this.indicesClient.getFieldMapping(any(), any())).thenReturn(getFieldMappingsResponse);
			XContentBuilder builder = XContentFactory.jsonBuilder();
			builder.startObject();
			builder.field("any field", new HashMap());
			builder.endObject();
			BytesReference bytesReference = BytesReference.bytes(builder);
			FieldMappingMetaData mappingMetaData = new FieldMappingMetaData(index, bytesReference);
			Map<String, FieldMappingMetaData> mapBuilder = new HashMap<>();
			mapBuilder.put("data.any field", mappingMetaData);
			Map<String, Map<String, FieldMappingMetaData>> mappingBuilder = new HashMap<>();
			mappingBuilder.put("any index 1", mapBuilder);
			mappingBuilder.put("any index 2", mapBuilder);
			Map<String, Map<String, Map<String, FieldMappingMetaData>>> mapping = new HashMap<>();
			mapping.put("indices 1", mappingBuilder);
			when(getFieldMappingsResponse.mappings()).thenReturn(mapping);
			doReturn(mappingResponse).when(this.indicesClient).putMapping(any(), any(RequestOptions.class));
			BulkByScrollResponse response = mock(BulkByScrollResponse.class);
			doReturn(response).when(this.restHighLevelClient).updateByQuery(any(), any(RequestOptions.class));
			when(response.getBulkFailures()).thenReturn(new ArrayList<Failure>());
			when(elasticClientHandler.createRestClient()).thenReturn(restHighLevelClient);
			this.sut.updateIndexMappingForIndicesOfSameType(indices,"any field invalid");
		} catch (Exception e) {
			throw e;
		}
	}

	@Test(expected = AppException.class)
	public void should_throw_exception_if_elastic_search_failedToFetch_andWeIndexfield_with_keyword() throws Exception {
		try {

			Set<String> indices = new HashSet<String>();
			indices.add("indices 1");
			indices.add("indices Invalid");
			GetFieldMappingsResponse getFieldMappingsResponse = mock(GetFieldMappingsResponse.class);
			doReturn(this.indicesClient).when(this.restHighLevelClient).indices();
			when(this.indicesClient.getFieldMapping(any(), any())).thenThrow(new ElasticsearchException(""));
			XContentBuilder builder = XContentFactory.jsonBuilder();
			builder.startObject();
			builder.field("any field", new HashMap());
			builder.endObject();
			BytesReference bytesReference = BytesReference.bytes(builder);
			FieldMappingMetaData mappingMetaData = new FieldMappingMetaData(index, bytesReference);
			Map<String, FieldMappingMetaData> mapBuilder = new HashMap<>();
			mapBuilder.put("data.any field", mappingMetaData);
			Map<String, Map<String, FieldMappingMetaData>> mappingBuilder = new HashMap<>();
			mappingBuilder.put("any index 1", mapBuilder);
			mappingBuilder.put("any index 2", mapBuilder);
			Map<String, Map<String, Map<String, FieldMappingMetaData>>> mapping = new HashMap<>();
			mapping.put("indices 1", mappingBuilder);
			when(getFieldMappingsResponse.mappings()).thenReturn(mapping);
			doReturn(mappingResponse).when(this.indicesClient).putMapping(any(), any(RequestOptions.class));
			BulkByScrollResponse response = mock(BulkByScrollResponse.class);
			doReturn(response).when(this.restHighLevelClient).updateByQuery(any(), any(RequestOptions.class));
			when(response.getBulkFailures()).thenReturn(new ArrayList<Failure>());
			when(elasticClientHandler.createRestClient()).thenReturn(restHighLevelClient);
			this.sut.updateIndexMappingForIndicesOfSameType(indices,"any field");
		} catch (AppException e) {
			throw e;
		}
	}

	@Test(expected = AppException.class)
	public void should_throw_exception_when_elastic_failedToIndex_indices_field_with_keyword() {
		try {
			Set<String> indices = new HashSet<String>();
			indices.add("indices 1");
			indices.add("indices Invalid");
			GetFieldMappingsResponse getFieldMappingsResponse = mock(GetFieldMappingsResponse.class);
			doReturn(this.indicesClient).when(this.restHighLevelClient).indices();
			when(this.indicesClient.getFieldMapping(any(), any())).thenReturn(getFieldMappingsResponse);
			XContentBuilder builder = XContentFactory.jsonBuilder();
			builder.startObject();
			builder.field("any field", new HashMap());
			builder.endObject();
			BytesReference bytesReference = BytesReference.bytes(builder);
			FieldMappingMetaData mappingMetaData = new FieldMappingMetaData(index, bytesReference);
			Map<String, FieldMappingMetaData> mapBuilder = new HashMap<>();
			mapBuilder.put("data.any field", mappingMetaData);
			Map<String, Map<String, FieldMappingMetaData>> mappingBuilder = new HashMap<>();
			mappingBuilder.put("any index 1", mapBuilder);
			mappingBuilder.put("any index 2", mapBuilder);
			Map<String, Map<String, Map<String, FieldMappingMetaData>>> mapping = new HashMap<>();
			mapping.put("indices 1", mappingBuilder);
			when(getFieldMappingsResponse.mappings()).thenReturn(mapping);
			doReturn(mappingResponse).when(this.indicesClient).putMapping(any(), any(RequestOptions.class));
			BulkByScrollResponse response = mock(BulkByScrollResponse.class);
			doReturn(response).when(this.restHighLevelClient).updateByQuery(any(), any(RequestOptions.class));
			when(response.getBulkFailures()).thenReturn(new ArrayList<Failure>());
			when(this.indicesClient.putMapping(any(), any(RequestOptions.class))).thenThrow(new ElasticsearchException(""));
			when(elasticClientHandler.createRestClient()).thenReturn(restHighLevelClient);
			this.sut.updateIndexMappingForIndicesOfSameType(indices,"any field");
		} catch (AppException e) {
			throw e;
		} catch (Exception e) {
			fail("Should not throw this exception" + e.getMessage());
		}
	}


	@Test
	public void should_returnDocumentMapping_givenValidIndexSchema() {

		try {
			Map<String, Object> documentMapping = this.sut.getIndexMappingFromRecordSchema(this.indexSchema);
			String documentMappingJson = new Gson().toJson(documentMapping);
			assertEquals(this.mappingValid, documentMappingJson);

		} catch (Exception e) {
			fail("Should not throw this exception" + e.getMessage());
		}
	}
}
